/*
 *  Copyright (C) Chaperon. All rights reserved.                               
 *  ------------------------------------------------------------------------- 
 *  This software is published under the terms of the Apache Software License 
 *  version 1.1, a copy of which has been included  with this distribution in 
 *  the LICENSE file.                                                         
 */

package net.sourceforge.chaperon.parser;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import net.sourceforge.chaperon.grammar.Grammar;

/**
 * This class will the parser produce, if an error occurs
 *
 * @author Stephan Michels
 * @version CVS $Id: ParserException.java,v 1.2 2002/05/02 16:08:57 benedikta Exp $
 */
public class ParserException extends Exception
{
  /** The URI of the grammar DTD */
  //public final static String URI = "-//Chaperon//DTD parser-exception 1.0//EN";
  
  /** Element name */ 
  //public final static String ELEMENT = "parser-exception";

  /** Attribute name of the message property */
  //public final static String MESSAGE_ATTRIBUTE = "message";

  /** Attribute name of the message property */
  //public final static String LINENUMBER_ATTRIBUTE = "linenr";

  /** Attribute name of the message property */
  //public final static String COLUMNNUMBER_ATTRIBUTE = "coumnnr";

  /** Element name for acceptedsymbols */ 
  //public final static String ACCEPTEDSYMBOL_ELEMENT = "parser-exception-accept";

  private String    _publicId = "";
  private String    _systemId = "";
  private String    _message = null;
  private int       _lineNumber = 0;
  private int       _columnNumber = 0;
  private Exception _exception = null;
  private String[]  _acceptedsymbols = new String[0];

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   */
  public ParserException(String message)
  {
    _message = message;
  }

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   * @param publicId Public id from the parsed document
   * @param systemId System if from the parsed document
   * @param lineNumber Line number in which the error occurs
   * @param columnNumber Column number in which the error occurs
   */
  public ParserException(String message, String publicId, String systemId,
                        int lineNumber, int columnNumber)
  {
    _message = message;
    _publicId = publicId;
    _systemId = systemId;
    _lineNumber = lineNumber;
    _columnNumber = columnNumber;
  }

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   * @param lineNumber Line number in which the error occurs
   * @param columnNumber Column number in which the error occurs
   */
  public ParserException(String message, int lineNumber, int columnNumber)
  {
    _message = message;
    _lineNumber = lineNumber;
    _columnNumber = columnNumber;
  }

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   * @param publicId Public id from the parsed document
   * @param systemId System if from the parsed document
   * @param lineNumber Line number in which the error occurs
   * @param columnNumber Column number in which the error occurs
   * @param e Additional exception
   */
  public ParserException(String message, String publicId, String systemId,
                        int lineNumber, int columnNumber, Exception e)
  {
    _message = message;
    _publicId = publicId;
    _systemId = systemId;
    _lineNumber = lineNumber;
    _columnNumber = columnNumber;
  }

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   * @param publicId Public id from the parsed document
   * @param systemId System if from the parsed document
   * @param lineNumber Line number in which the error occurs
   * @param columnNumber Column number in which the error occurs
   * @param acceptedsymbols Which symbols the parser at this position accepted
   */
  public ParserException(String message, String publicId, String systemId,
                        int lineNumber, int columnNumber,
                        String[] acceptedsymbols)
  {
    _message = message;
    _publicId = publicId;
    _systemId = systemId;
    _lineNumber = lineNumber;
    _columnNumber = columnNumber;
    _acceptedsymbols = acceptedsymbols;
  }

  /**
   * Constructs a exception
   *
   * @param message Message of the error
   * @param lineNumber Line number in which the error occurs
   * @param columnNumber Column number in which the error occurs
   * @param acceptedsymbols Which symbols the parser at this position accepted
   */
  public ParserException(String message,
                        int lineNumber, int columnNumber,
                        String[] acceptedsymbols)
  {
    _message = message;
    _lineNumber = lineNumber;
    _columnNumber = columnNumber;
    _acceptedsymbols = acceptedsymbols;
  }

  /**
   * Return the column number in which the error occurs
   *
   * @return Column number in which the error occurs
   */
  public int getColumnNumber()
  {
    return _columnNumber;
  }

  /**
   * Returns the line number in which the error occurs
   *
   * @return Line number in which the error occurs
   */
  public int getLineNumber()
  {
    return _lineNumber;
  }

  /**
   * Returns the public id from the parsed document
   *
   * @return Public id from the parsed document
   */
  public String getPublicId()
  {
    return _publicId;
  }

  /**
   * Returns the system if from the parsed document
   *
   * @return System if from the parsed document
   */
  public String getSystemId()
  {
    return _systemId;
  }

  /**
   * Returns the symbols, which the parser at this position accepted
   *
   * @return Which symbols the parser at this position accepted
   */
  public String[] getAcceptedSymbols()
  {
    return _acceptedsymbols;
  }

  /**
   * Return the additional exception
   *
   * @return Additional exception
   */
  public Exception getException()
  {
    return _exception;
  }

  /**
   * Return the message of the error
   *
   * @return Message of the error
   */
  public String getMessage()
  {

    if (_message != null)
      return _message;
    else if (_exception != null)
      return _exception.getMessage();
    return "";
  }

  /**
   * Return the message of the error
   *
   * @return Message of the error
   */
  public String toString()
  {
    StringBuffer buffer = new StringBuffer();

    buffer.append(getMessage());
    buffer.append("\n");
    buffer.append("Syntax error near line ");
    buffer.append(String.valueOf(_lineNumber));
    buffer.append("/column ");
    buffer.append(String.valueOf(_columnNumber));
    buffer.append("\nexpected ");
    for (int i = 0; i < _acceptedsymbols.length; i++)
    {
      if (i > 0)
        buffer.append(",");
      buffer.append(_acceptedsymbols[i]);
    }
    buffer.append("\n");

    return buffer.toString();
  }

  /**
   * Serialize the exception to a SAX stream
   *
   * @param handler The content handler, which receives the SAX events
   *
   * @throws SAXException
   */
  /*public void toSAX(ContentHandler handler) throws SAXException
  {
    handler.startDocument();

    AttributesImpl attributes = new AttributesImpl();
    
    attributes.addAttribute(null, MESSAGE_ATTRIBUTE, MESSAGE_ATTRIBUTE,
                            "CDATA", message);
    attributes.addAttribute(null, LINENUMBER_ATTRIBUTE, LINENUMBER_ATTRIBUTE,
                            "CDATA", String.valueOf(lineNumber));
    attributes.addAttribute(null, COLUMNNUMBER_ATTRIBUTE, COLUMNNUMBER_ATTRIBUTE,
                            "CDATA", String.valueOf(columnNumber));

    handler.startElement(URI, ELEMENT, ELEMENT, attributes);

    for(int i=0; i<acceptedsymbols.length; i++)
    {
      handler.startElement(URI, ACCEPTEDSYMBOL_ELEMENT, ACCEPTEDSYMBOL_ELEMENT, new AttributesImpl());
      handler.characters(acceptedsymbols[i].toCharArray(), 0, acceptedsymbols[i].length());
      handler.endElement(URI, ACCEPTEDSYMBOL_ELEMENT, ACCEPTEDSYMBOL_ELEMENT);
    } 

    handler.endElement(URI, ELEMENT, ELEMENT);

    handler.endDocument();
  }*/
}
